{
 *****************************************************************************
 *                                                                           *
 *  See the file COPYING.modifiedLGPL.txt, included in this distribution,    *
 *  for details about the copyright.                                         *
 *                                                                           *
 *  This program is distributed in the hope that it will be useful,          *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                     *
 *                                                                           *
 *****************************************************************************
Filename:	oodrawwriter.pas
Use:		Write OpenOffice Draw (odg) Data from Canvas procedures
Author: 	Dr. Marcus Sackrow
Date:		July 2009
}
unit oodrawwriter;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, XMLWrite, DOM, Zip, ZipUtils, Graphics, FileUtil,
  {$IFDEF WINDOWS}
  Windows,
  {$ENDIF}
  Math, GraphMath;

type       Todg=class(Tcanvas)
                 private
                   FHandle: THandle;
                   FPenPos : TPoint;
                   unique:integer;  // Unique number for styles and files
                   isgroup:boolean; // if true = inside a group (only one level allowed up to now)

                   FWidth:integer;
                   FHeight:integer;

                   Content:TXMLDocument;   //Root for Content.xml
                   Manifest:TXMLDocument;  //Root for Manifest.xml
                   ManBase:TDOMNode;       //base for add new files to manifest/zip
                   ConStyle:TDOMNode;      //base for add new styles
                   ConBody:TDOMNode;       //base for add new items in the draw
                   ConSave:TDOMNode;       //save of the base ConBody
                                           // when grouping is on the Conbody
                                           // point to the Group item

                   PicList:TStringList; // Pictures in the file
                   TEMPDIR:string;      // for the files before put to zip (init at create)
                   procedure MakeBasicManifest;   //make the basic layout
                   procedure MakeBasicContent;
                   function MakeStyle:string;
                   function MakeFontStyle:string;
                   function GetRGBColor(Color:Tcolor):string;
                   procedure CheckSize(ax,ay:integer); //set width and height
                   procedure SetPixel(X,Y: Integer; Value: TColor); override; //for property Pixels
                 public
                   //class function
                   constructor Create;
                   destructor Destroy; override;

                   // ODG Functions
                   procedure GroupStart;
                   procedure GroupEnd;
                   procedure WriteODG(filename:string);

                   //Move to/Clear
                   procedure MoveTo(X1,Y1:integer); override;
                   procedure MoveTo(p:TPoint);
                   procedure Clear;

                   // ## Draw functions ##
                   // line
                   procedure Line(x1,y1,x2,y2:integer); override;
                   procedure Line(const p1,p2:TPoint);
                   procedure Line(const Points:Trect);
                   procedure LineTo(X1,Y1:integer); override;
                   procedure Polyline(Points: PPoint; NumPts: Integer); override;


                   // rect
                   procedure Rectangle(x1,y1,x2,y2:integer); override;
                   procedure Rectangle(const Arect:Trect);
                   procedure FillRect(const ARect: TRect); override;
                   procedure FillRect(X1,Y1,X2,Y2: Integer);


                   // ellipse
                   procedure Ellipse(x1,y1,x2,y2:integer);  override;
                   procedure Ellipse(const ARect:Trect);
                   procedure EllipseC(x,y:integer; rx,ry:longword);

                   // Polygon
                   procedure Polygon(Points: PPoint; NumPts: Integer; Winding: boolean = False); override;

                   // Arc
                   procedure Arc(ALeft, ATop, ARight, ABottom, angle1, angle2: Integer); override;
                   // Chord
                   procedure Chord(x1, y1, x2, y2, StartAngle16Deg, EndAngle16Deg: Integer); override;
                   // Radial Pie
                   procedure RadialPie(x1, y1, x2, y2, StartAngle16Deg, EndAngle16Deg: Integer); override;

                   // Textfield
                   procedure TextOut(X,Y:Integer;const Text:String); override;

                   // Pixels
                   property Pixels[X, Y: Integer]: TColor write SetPixel;

                   //stretch Draw
                   procedure StretchDraw(const DestRect: TRect; SrcGraphic: TGraphic); override;
                   // Draw
                   procedure Draw(X,Y: Integer; SrcGraphic: TGraphic); override;

                   // properties
                   property Handle:Thandle read FHandle write FHandle;
                   property width:integer read FWidth;
                   property height:integer read FHeight;
                 end;

implementation

//###########################################################################
// Constructor

constructor Todg.create;
var dir: array [0..256] of Char;
Begin
    // Creates
  inherited create;
  content:=TXMLDocument.create;
  manifest:=TXMLDocument.create;
  PicList:=TStringList.Create;
    // make basic layout
  makebasicmanifest;
  makebasiccontent;
    // init some variables
  unique:=0;
  FPenPos.x:=0;
  FPenPos.y:=0;
  FWidth:=0;
  FHeight:=0;
  isgroup:=false;
    // Set TEMPDIR for the files before copied to the ODG/ZIP file
{$IFDEF WINDOWS}
  windows.GetWindowsDirectory(dir, 256);
  TEMPDIR:=StrPas(dir)+DIRECTORYSEPARATOR+'temp'+DIRECTORYSEPARATOR;
{$ENDIF}
{$IFDEF UNIX}
  TEMPDIR:=DIRECTORYSEPARATOR+'tmp'+DIRECTORYSEPARATOR;
{$ENDIF}
end;


//###########################################################################
// Destructor

destructor Todg.destroy;
var i:integer;
Begin
  content.Free;
  manifest.free;
  for i:=0 to Piclist.Count-1 do
  Begin
    TPortableNetWorkGraphic(Piclist.Objects[i]).free;
  end;
  PicList.Free;
  inherited;
end;

//###########################################################################
// Clear

Procedure Todg.Clear;
var i:integer;
Begin
  content.Free;
  manifest.free;
  for i:=0 to Piclist.Count-1 do
  Begin
    TPortableNetWorkGraphic(Piclist.Objects[i]).free;
  end;
  PicList.Free;
  content:=TXMLDocument.create;
  manifest:=TXMLDocument.create;
  PicList:=TStringList.Create;
  makebasicmanifest;
  makebasiccontent;
  unique:=0;
  FPenPos.x:=0;
  FPenPos.y:=0;
  FWidth:=0;
  FHeight:=0;
  isgroup:=false;
end;

//###########################################################################
//make basic Content

procedure Todg.makebasiccontent;
var item,item0:TDOMNode;
Begin
  with content do
  Begin
    XMLVersion:='1.0';
    Encoding:='UTF-8';

 // define namespace
    item0:= CreateElement('office:document-content');
      TDOMElement(item0).SetAttribute('xmlns:draw', 'urn:oasis:names:tc:opendocument:xmlns:drawing:1.0');
      TDOMElement(item0).SetAttribute('xmlns:office', 'urn:oasis:names:tc:opendocument:xmlns:office:1.0');
      TDOMElement(item0).SetAttribute('xmlns:svg', 'urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0');
      TDOMElement(item0).SetAttribute('xmlns:xlink', 'http://www.w3.org/1999/xlink');
      TDOMElement(item0).SetAttribute('xmlns:style','urn:oasis:names:tc:opendocument:xmlns:style:1.0');
      TDOMElement(item0).SetAttribute('xmlns:text','urn:oasis:names:tc:opendocument:xmlns:text:1.0');
      TDOMElement(item0).SetAttribute('xmlns:fo','urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0');
      TDOMElement(item0).SetAttribute('office:version', '1.2');
    content.AppendChild(item0);
 //create style template
    constyle := CreateElement('office:automatic-styles');
    item0.AppendChild(constyle);

 //Body for drawings
    item := CreateElement('office:body');
    item0.appendchild(item);
    conbody := CreateElement('office:drawing');
    item.AppendChild(conbody);
    item := CreateElement('draw:page');
    conbody.AppendChild(item);
    conbody:=item;
  end;
end;

//###########################################################################
//make basic Manifest;

procedure Todg.makebasicmanifest;
var item:TDOMNode;
Begin
  with manifest do
  Begin
    XMLVersion:='1.0';
    Encoding:='UTF-8';
    // namespace
    manbase := CreateElement('manifest:manifest');
      TDOMElement(manbase).SetAttribute('xmlns:manifest', 'urn:oasis:names:tc:opendocument:xmlns:manifest:1.0');
    manifest.AppendChild(manbase);
    //base file entry
    item := CreateElement('manifest:file-entry');
      TDOMElement(item).SetAttribute('manifest:media-type', 'application/vnd.oasis.opendocument.graphics');
      TDOMElement(item).SetAttribute('manifest:version', '1.2');
      TDOMElement(item).SetAttribute('manifest:full-path', '/');
    manbase.AppendChild(item);
    // include the content.xml, always needed so already put in
    item := CreateElement('manifest:file-entry');
      TDOMElement(item).SetAttribute('manifest:media-type', 'text/xml');
      TDOMElement(item).SetAttribute('manifest:full-path', 'content.xml');
    manbase.AppendChild(item);
  end;
end;

//###########################################################################
// Start a group

procedure Todg.GroupStart;
Begin
  if not isgroup then
  Begin
    ConSave:=ConBody;
    ConBody := content.CreateElement('draw:g');
    ConSave.AppendChild(ConBody);
    isgroup:=true;
  end;
end;

//###########################################################################
// End of Group

procedure Todg.GroupEnd;
Begin
  if isgroup then
  Begin
    ConBody:=ConSave;
    isgroup:=false;
  end;
end;

//###########################################################################
// Make style Information for a text (Font included)

function Todg.MakeFontStyle:string;
var item, item2:TDOMNode;
    sname:string;
Begin
  unique:=unique+1;
  sname:='text'+inttostr(unique);   // "text"+number = style name
  Result:=sname;
  item := content.CreateElement('style:style');
    TDOMElement(item).SetAttribute('style:name', sname); // set the style name
    TDOMElement(item).SetAttribute('style:family', 'graphic');
  constyle.AppendChild(item);
  item2 := content.CreateElement('style:graphic-properties');
    TDOMElement(item2).SetAttribute('draw:stroke','none');    // turn off border
    TDOMElement(item2).SetAttribute('draw:fill', 'none');     // no fill
    TDOMElement(item2).SetAttribute('draw:auto-grow-height', 'true'); // let it grow self then width is not important ;)
    TDOMElement(item2).SetAttribute('draw:auto-grow-width', 'true');
  item.AppendChild(item2);

   //style for Font
  item := content.CreateElement('style:style');
    TDOMElement(item).SetAttribute('style:name', 'font'+sname);  //'Fonttext'+number = style for text
    TDOMElement(item).SetAttribute('style:family', 'paragraph'); // complete textfield is one font
  constyle.AppendChild(item);
  item2 := content.CreateElement('style:text-properties');
    TDOMElement(item2).SetAttribute('fo:font-family',Font.Name);               // font name
    TDOMElement(item2).SetAttribute('fo:font-size', inttostr(font.Size)+'pt'); // size
    TDOMElement(item2).SetAttribute('fo:color', GetRGBColor(Font.Color));      // color
  item.AppendChild(item2);
end;

//###########################################################################
// Tcolor to HexColor-String

function Todg.GetRGBColor(Color:Tcolor):string;
Begin
  Result:='#'+
  IntToHex(red(Color),2)+
  IntToHex(green(Color),2)+
  IntToHex(blue(Color),2);
end;

//###########################################################################
// Make Style informations for normal Graphics

function Todg.MakeStyle:string;
var item, item2:TDOMNode;
    sname:string;
Begin
  unique:=unique+1;
  sname:='form'+inttostr(unique);   // "Form" + number = style for graphics
  Result:=sname;                    // give back to calling
  item := content.CreateElement('style:style');
    TDOMElement(item).SetAttribute('style:name', sname);        // set the style name
    TDOMElement(item).SetAttribute('style:family', 'graphic');  // graphic style
  constyle.AppendChild(item);
  item2 := content.CreateElement('style:graphic-properties');
    if Pen.Style<>psSolid then                                  // set Dash if not a solid line
    Begin
      TDOMElement(item2).SetAttribute('draw:stroke', 'dash');
      case Pen.Style of
         psDot:TDOMElement(item2).SetAttribute('draw:stroke-dash', 'Fine_20_Dotted');
         psDashDot:TDOMElement(item2).SetAttribute('draw:stroke-dash', 'Line_20_with_20_Fine_20_Dots');
         psDashDotDot:TDOMElement(item2).SetAttribute('draw:stroke-dash', 'Ultrafine_20_2_20_Dots_20_3_20_Dashes');
         psinsideFrame:TDOMElement(item2).SetAttribute('draw:stroke-dash', 'Ultrafine_20_Dashed');
         // more to come
         psClear:TDOMElement(item2).SetAttribute('draw:stroke','none');   // no line
      end;
    end;
    TDOMElement(item2).SetAttribute('svg:stroke-color', GetRGbColor(Pen.Color));  // color
    TDOMElement(item2).SetAttribute('svg:stroke-width', inttostr(Pen.Width)+'pt'); //width
    TDOMElement(item2).SetAttribute('draw:fill-color', GetRGBColor(Brush.Color));  // fill color
    case Brush.Style of
         bsSolid:TDOMElement(item2).SetAttribute('draw:fill', 'solid');            // fill style
         //More to come
         bsClear:TDOMElement(item2).SetAttribute('draw:fill', 'none');
    end;

  item.AppendChild(item2);
end;

//###########################################################################
// MoveTo (Move Pen Position)

procedure Todg.MoveTo(X1,Y1:integer);
Begin
  FPenPos.x:=x1;
  FPenPos.y:=y1;
end;

procedure Todg.MoveTo(p:TPoint);
Begin
  FPenPos:=p;
end;



// ##########################################################################
// line (x1,y1,x2,y2)
// line(p1,p2) Line(TRect)

procedure Todg.line(x1,y1,x2,y2:integer);
var item:TDOMNode;
    sname:string;
Begin
    // if bigger than width and height -> extend them
  checksize(x1,y1);
  checksize(x2,y2);
    // get/make a style for this object
  sname:=makestyle;
    // set the parameter
  item := content.CreateElement('draw:line');
    TDOMElement(item).SetAttribute('draw:style-name', sname);
    TDOMElement(item).SetAttribute('svg:x1', floattostrf(x1,fffixed,8,2)+'pt');
    TDOMElement(item).SetAttribute('svg:y1', floattostrf(y1,fffixed,8,2)+'pt');
    TDOMElement(item).SetAttribute('svg:x2', floattostrf(x2,fffixed,8,2)+'pt');
    TDOMElement(item).SetAttribute('svg:y2', floattostrf(y2,fffixed,8,2)+'pt');
  conbody.AppendChild(item);
    // move the pen position to the end point
  MoveTo(x2,y2);
end;

procedure Todg.Line(const p1,p2:Tpoint);
Begin
  line(p1.x,p1.y,p2.x,p2.y);
end;

procedure Todg.Line(const Points:Trect);
Begin
  line(Points.Left,Points.Top,Points.Right,Points.Bottom);
end;

procedure Todg.LineTo(X1,Y1:integer);
Begin
  line(FPenPos.x,FPenPos.y,x1,y1);
end;

procedure Todg.Polyline(Points: PPoint; NumPts: Integer);
var i:integer;
Begin
  Moveto(Points[0].x,Points[0].y);
  for i:=1 to Numpts-1 do
  Begin
    LineTo(Points[i].x,Points[i].y);
  end;
end;

// ##########################################################################
// rectangle (x1,y1,x2,y2)
// line(p1,p2) Line(TRect)

procedure Todg.Rectangle(x1,y1,x2,y2:integer);
var item:TDOMNode;
    sname:string;
Begin

  checksize(x1,y1);
  checksize(x2,y2);

  sname:=makestyle;

  item := content.CreateElement('draw:rect');
    TDOMElement(item).SetAttribute('draw:style-name', sname);
    TDOMElement(item).SetAttribute('svg:x', floattostrf(x1,fffixed,8,2)+'pt');
    TDOMElement(item).SetAttribute('svg:y', floattostrf(y1,fffixed,8,2)+'pt');
    TDOMElement(item).SetAttribute('svg:width', floattostrf(x2-x1,fffixed,8,2)+'pt');
    TDOMElement(item).SetAttribute('svg:height', floattostrf(y2-y1,fffixed,8,2)+'pt');
  conbody.AppendChild(item);

  MoveTo(x2,y2);
end;

procedure Todg.Rectangle(const Arect:Trect);
Begin
  Rectangle(ARect.Left,ARect.Top,ARect.Right,ARect.Bottom);
end;

procedure Todg.FillRect(const ARect: TRect);
Begin
   Rectangle(ARect.Left,ARect.Top,Arect.Right,ARect.Bottom);
end;

procedure Todg.FillRect(X1,Y1,X2,Y2: Integer);
Begin
  Rectangle(x1,y1,x2,y2);
end;


// #####################################################################
// ellipse

procedure Todg.Ellipse(x1,y1,x2,y2:integer);
var item:TDOMNode;
    sname:string;
Begin

  checksize(x1,y1);
  checksize(x2,y2);

  sname:=makestyle;

  item := content.CreateElement('draw:ellipse');
    TDOMElement(item).SetAttribute('draw:style-name', sname);
    TDOMElement(item).SetAttribute('svg:x', inttostr(x1)+'pt');
    TDOMElement(item).SetAttribute('svg:y', inttostr(y1)+'pt');
    TDOMElement(item).SetAttribute('svg:width', inttostr(x2-x1)+'pt');
    TDOMElement(item).SetAttribute('svg:height', inttostr(y2-y1)+'pt');
  conbody.AppendChild(item);

  MoveTo(x2,y2);
end;

procedure Todg.Ellipse(const ARect:Trect);
Begin
  Ellipse(ARect.Left,ARect.Top,ARect.Right,ARect.Bottom);
end;

procedure Todg.EllipseC(x,y:integer;rx,ry:longword);
begin
  Ellipse (x-rx,y-ry,x+rx,y+ry);
end;

// #####################################################################
// Polygon

procedure Todg.Polygon(Points: PPoint; NumPts: Integer; Winding: boolean = False);
  var item:TDOMNode;
    sname:string;
    coords:string;
    i:integer;
    x1,x2,y1,y2:integer;
Begin
  coords:='';
  x1:=99999;
  y1:=99999;
  x2:=0;
  y2:=0;

  // get max/min for x and y for the Viewbox and rel. coordinates
  for i:=0 to Numpts-1 do
  Begin
    checksize(Points[i].x,Points[i].y);
    x1:=min(Points[i].x,x1);
    x2:=max(Points[i].x,x2);
    y1:=min(Points[i].y,y1);
    y2:=max(Points[i].y,y2);
  end;

  // make the coords for the Polygon

  for i:=0 to Numpts-1 do
  Begin
    coords:=coords+' '+inttostr(Points[i].x-x1)+','+inttostr(Points[i].y-y1);
  end;

// get the style mark
  sname:=makestyle;

// make the entries
  item := content.CreateElement('draw:polygon');
    TDOMElement(item).SetAttribute('draw:style-name', sname);
    TDOMElement(item).SetAttribute('svg:x', inttostr(x1)+'pt');
    TDOMElement(item).SetAttribute('svg:y', inttostr(y1)+'pt');
    TDOMElement(item).SetAttribute('svg:width', inttostr((x2-x1))+'pt');
    TDOMElement(item).SetAttribute('svg:height', inttostr((y2-y1))+'pt');
    TDOMElement(item).SetAttribute('svg:viewBox', inttostr(0)+' '+inttostr(0)+' '+inttostr(x2-x1)+' '+inttostr(y2-y1));
    TDOMElement(item).SetAttribute('draw:points', coords);
  conbody.AppendChild(item);

// move to last point for further painting with lineto
  MoveTo(Points[Numpts-1].x,Points[Numpts-1].y);
end;


// #####################################################################
// Arc  angle[1,2] = angle[1,2]*16 :O? who make such stupid thing?

procedure Todg.Arc(ALeft, ATop, ARight, ABottom, angle1, angle2: Integer);
var item:TDOMNode;
    sname:string;
    x1,x2,y1,y2:integer;
Begin
  x1:=ALeft;
  x2:=ARight;
  y1:=ATop;
  y2:=ABottom;

  checksize(x1,y1);
  checksize(x2,y2);

  sname:=makestyle;

  item := content.CreateElement('draw:ellipse');
    TDOMElement(item).SetAttribute('draw:style-name', sname);
    TDOMElement(item).SetAttribute('svg:x', inttostr(x1)+'pt');
    TDOMElement(item).SetAttribute('svg:y', inttostr(y1)+'pt');
    TDOMElement(item).SetAttribute('svg:width', inttostr(x2-x1)+'pt');
    TDOMElement(item).SetAttribute('svg:height', inttostr(y2-y1)+'pt');
    TDOMElement(item).SetAttribute('draw:kind', 'arc');
    TDOMElement(item).SetAttribute('draw:start-angle', floattostrf(angle1/16,fffixed,8,0));
    TDOMElement(item).SetAttribute('draw:end-angle', floattostrf(angle2/16,fffixed,8,0));
  conbody.AppendChild(item);

  MoveTo(x2,y2);
end;

// #####################################################################
// Chord  angle = angle*16 :O? who make such stupid thing?

procedure Todg.Chord(x1, y1, x2, y2, StartAngle16Deg, EndAngle16Deg: Integer);
var item:TDOMNode;
    sname:string;
Begin

  checksize(x1,y1);
  checksize(x2,y2);

  sname:=makestyle;

  item := content.CreateElement('draw:ellipse');
    TDOMElement(item).SetAttribute('draw:style-name', sname);
    TDOMElement(item).SetAttribute('svg:x', inttostr(x1)+'pt');
    TDOMElement(item).SetAttribute('svg:y', inttostr(y1)+'pt');
    TDOMElement(item).SetAttribute('svg:width', inttostr(x2-x1)+'pt');
    TDOMElement(item).SetAttribute('svg:height', inttostr(y2-y1)+'pt');
    TDOMElement(item).SetAttribute('draw:kind', 'cut');
    TDOMElement(item).SetAttribute('draw:start-angle', floattostrf(StartAngle16Deg/16,fffixed,8,0));
    TDOMElement(item).SetAttribute('draw:end-angle', floattostrf(EndAngle16Deg/16,fffixed,8,0));
  conbody.AppendChild(item);

  MoveTo(x2,y2);
end;

// #####################################################################
// RadialPie  angle = angle*16 :O? who make such stupid thing?

procedure Todg.RadialPie(x1, y1, x2, y2, StartAngle16Deg, EndAngle16Deg: Integer);
var item:TDOMNode;
    sname:string;
Begin

  checksize(x1,y1);
  checksize(x2,y2);

  sname:=makestyle;

  item := content.CreateElement('draw:ellipse');
    TDOMElement(item).SetAttribute('draw:style-name', sname);
    TDOMElement(item).SetAttribute('svg:x', inttostr(x1)+'pt');
    TDOMElement(item).SetAttribute('svg:y', inttostr(y1)+'pt');
    TDOMElement(item).SetAttribute('svg:width', inttostr(x2-x1)+'pt');
    TDOMElement(item).SetAttribute('svg:height', inttostr(y2-y1)+'pt');
    TDOMElement(item).SetAttribute('draw:kind', 'section');
    TDOMElement(item).SetAttribute('draw:start-angle', floattostrf(StartAngle16Deg/16,fffixed,8,0));
    TDOMElement(item).SetAttribute('draw:end-angle', floattostrf(EndAngle16Deg,fffixed,8,0));
  conbody.AppendChild(item);

  MoveTo(x2,y2);
end;


// #####################################################################
// Textout

procedure Todg.TextOut(X,Y:Integer;const Text:String);
var item,item2:TDOMNode;
    sname:string;
    x2,y2:integer;
Begin

  checksize(x,y);
  y2:=y+10;
  x2:=x+10;

  checksize(x2,y2);

// if font not set ty to fix it
  if font.size=0 then font.size:=11;
  if font.Name='' then font.Name:='Arial';

// make th font style (incomplete only size and name are set)
  sname:=MakeFontStyle;

// first make a bounding box, width, height not important, as long as
// automatical extent is activated see Fontstyle

  item := content.CreateElement('draw:frame');
    TDOMElement(item).SetAttribute('draw:style-name', sname);
    TDOMElement(item).SetAttribute('svg:x', inttostr(x)+'pt');
    TDOMElement(item).SetAttribute('svg:y', inttostr(y)+'pt');
    TDOMElement(item).SetAttribute('svg:width', inttostr(x2-x)+'pt');
    TDOMElement(item).SetAttribute('svg:height', inttostr(y2-y)+'pt');
  conbody.AppendChild(item);
  item2 := content.CreateElement('draw:text-box');
  item.AppendChild(item2);
// write the text self, 'font'+sname = style text style belongs to the graphics style sname;
  item := content.CreateElement('text:p');
    TDOMElement(item).SetAttribute('text:style-name', 'font'+sname);
    TDOMElement(item).TextContent:=Text;
  item2.AppendChild(item);


  MoveTo(x2,y2);
end;


// #####################################################################
// Set a Pixel (makes a strange work-around ;), but it works :P)

procedure Todg.SetPixel(X,Y: Integer; Value: TColor);
Begin
  Brush.Color:=Value;
  Pen.Style:=psClear;
  Rectangle(x,y,x+1,y+1);
end;

// #####################################################################
// StretchDraw

procedure Todg.StretchDraw(const DestRect: TRect; SrcGraphic: TGraphic);
var Bitmap:TPortableNetworkGraphic;
    x1,y1:integer;
    filename:string;
    item,item2:TDOMNode;
    x2,y2:integer;
    nRect:TRect;
Begin
  Bitmap:=TPortableNetworkGraphic.Create;
  nRect.Top:=0;
  nRect.Left:=0;
  nRect.Right:=DestRect.Right-DestRect.left;
  nRect.Bottom:=DestRect.Bottom-DestRect.Top;
  Bitmap.SetSize(nRect.Right,nRect.Bottom);
  Bitmap.Canvas.StretchDraw(nRect,SrcGraphic);
  unique:=unique+1;
  filename:='image'+inttostr(unique)+'.png';
  Piclist.AddObject(filename,Bitmap);

  x1:=DestRect.Right;
  x2:=DestRect.left;
  y1:=DestRect.Top;
  y2:=DestRect.Bottom;

  checksize(x2,y2);


// first make a bounding box
// set the image into

  item := content.CreateElement('draw:frame');
    //TDOMElement(item).SetAttribute('draw:style-name', sname);
    TDOMElement(item).SetAttribute('svg:x', inttostr(x1)+'pt');
    TDOMElement(item).SetAttribute('svg:y', inttostr(y1)+'pt');
    TDOMElement(item).SetAttribute('svg:width', inttostr(x2-x1)+'pt');
    TDOMElement(item).SetAttribute('svg:height', inttostr(y2-y1)+'pt');
  conbody.AppendChild(item);
  item2 := content.CreateElement('draw:image');
    TDOMElement(item2).SetAttribute('xlink:href', 'Pictures/'+filename);
    TDOMElement(item2).SetAttribute('xlink:type', 'simple');
    TDOMElement(item2).SetAttribute('xlink:show', 'embed');
    TDOMElement(item2).SetAttribute('xlink:actuate', 'onLoad');
  item.AppendChild(item2);
//make entry in Manifest for the file (works without in Oo.org, but better make it)
  item := manifest.CreateElement('manifest:file-entry');
      TDOMElement(item).SetAttribute('manifest:media-type', 'image/png');
      TDOMElement(item).SetAttribute('manifest:full-path', 'Pictures/'+filename);
  manbase.AppendChild(item);

  MoveTo(x2,y2);
end;

// #####################################################################
// Draw

procedure Todg.Draw(X,Y: Integer; SrcGraphic: TGraphic);
var Bitmap:TPortableNetworkGraphic;
    x1,y1:integer;
    filename:string;
    item,item2:TDOMNode;
    x2,y2:integer;
    nRect:TRect;
Begin
  Bitmap:=TPortableNetworkGraphic.Create;
  nRect.Top:=0;
  nRect.Left:=0;
  nRect.Right:=SrcGraphic.Width;
  nRect.Bottom:=SrcGraphic.Height;
  Bitmap.SetSize(nRect.Right,nRect.Bottom);
  Bitmap.Canvas.StretchDraw(nRect,SrcGraphic);
  unique:=unique+1;
  filename:='image'+inttostr(unique)+'.png';
  Piclist.AddObject(filename,Bitmap);

  x1:=x;
  x2:=x+SrcGraphic.Width;
  y1:=y;
  y2:=y+SrcGraphic.Height;

  checksize(x2,y2);


// first make a bounding box
// set the image into

  item := content.CreateElement('draw:frame');
    //TDOMElement(item).SetAttribute('draw:style-name', sname);
    TDOMElement(item).SetAttribute('svg:x', inttostr(x1)+'pt');
    TDOMElement(item).SetAttribute('svg:y', inttostr(y1)+'pt');
    TDOMElement(item).SetAttribute('svg:width', inttostr(x2-x1)+'pt');
    TDOMElement(item).SetAttribute('svg:height', inttostr(y2-y1)+'pt');
  conbody.AppendChild(item);
  item2 := content.CreateElement('draw:image');
    TDOMElement(item2).SetAttribute('xlink:href', 'Pictures/'+filename);
    TDOMElement(item2).SetAttribute('xlink:type', 'simple');
    TDOMElement(item2).SetAttribute('xlink:show', 'embed');
    TDOMElement(item2).SetAttribute('xlink:actuate', 'onLoad');
  item.AppendChild(item2);
//make entry in Manifest (works without in OO.org, but better make it)
  item := manifest.CreateElement('manifest:file-entry');
      TDOMElement(item).SetAttribute('manifest:media-type', 'image/png');
      TDOMElement(item).SetAttribute('manifest:full-path', 'Pictures/'+filename);
  manbase.AppendChild(item);

  MoveTo(x2,y2);
end;

// #####################################################################
//write odgfile

procedure Todg.WriteODG(filename:string);
  var azip:zipfile;
    fzip:zip_fileinfo;
    i:integer;

 // subroutine for save a file to the ZIP
 procedure savetozip(fromf,tof:string);
 var results:integer;
     buffer:pointer;
     F:file;
 Begin
   zipOpenNewFileInZip(azip, Pchar(tof), @fzip, NIL,0 ,NIL,0,'file',0,9);
   assignfile(F,fromf);
   reset(F,1);
   getmem(buffer,1024);   // get mem
   results:=0;
   while not EOF(F) do    // copy file
   Begin
     blockread(f,buffer^,1024,results);
     zipWriteInFileInZip(azip, buffer, results);
   end;
   freemem(buffer,1024);   // free mem
   closefile(F);           // close file
   zipCloseFileInZip(azip);
 end;

Begin
  // save the xmls to Tempdir
  writexmlfile(manifest,TEMPDIR+'manifest.xml');
  writexmlfile(content,TEMPDIR+'content.xml');

  // i doenst care about the dates in the file :P
  fzip.dosDate:=0;
  fzip.tmz_date.tm_year:=2009;
  fzip.tmz_date.tm_mon:=01;
  fzip.tmz_date.tm_mday:=01;

  // ??? how to set to what? seems no problem if 0
  fzip.external_fa:=0;
  fzip.internal_fa:=0;

  //Open the odg file
  azip:=zipopen(PChar(utf8tosys(filename)),0);

  //Content write to the odg, remove tempfile
  savetozip(TEMPDIR+'content.xml','content.xml');
  deletefile(PChar(TEMPDIR+'content.xml'));

  //Manifest write to the odg, remove tempfile
  savetozip(TEMPDIR+'manifest.xml','META-INF/manifest.xml');
  deletefile(PChar(TEMPDIR+'manifest.xml'));

  //Pictures
  for i:=0 to PicList.Count-1 do
  Begin
    // save the Picture
    TPortableNetworkGraphic(Piclist.Objects[i]).SaveToFile(TEMPDIR+Piclist.Strings[i]);
    // copy to ODG
    savetozip(TEMPDIR+Piclist.Strings[i],'Pictures/'+Piclist.Strings[i]);
    // remove file
    deletefile(PChar(TEMPDIR+Piclist.Strings[i]));
  end;
  // finish
  zipclose(azip,'end');
end;

// set width and height to the max used values
procedure Todg.CheckSize(ax,ay:integer);
Begin
  Fwidth:=max(Fwidth,ax);
  FHeight:=max(FHeight,ay);
end;

end.

